/*** GRAFICOS.C ***/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <conio.h>
#include <graphics.h>
#include <dos.h>
#include <string.h>
#include <dir.h>

float dist_obs  = 10000,  /* Distancia do observador `a origem               */
      dist_plano = 2000,  /* Distancia do plano de projecao `a origem        */
      a, b, c, d,         /* Equacao do plano de projecao: ax+by+cz+d = 0    */
      u1, u2, u3,         /* Vetor 1 tangente ao plano de projecao           */
      v1, v2, v3,         /* Vetor 2 tangente ao plano de projecao           */
      A, B,                       /* Coordenadas do centro da tela           */
      TAM = 20,                   /* Coeficiente de ampliacao                */
      n1, n2, n3,                 /* Vetor normal                            */
      x0, y0, z0,                 /* Posicao do observador                   */
      nu = 40, nv = 40,           /* Subidivisoes do dominio                 */
      inc = 0.02,                 /* Incremento do parametro                 */
      P1, P2, P3,                 /* Ponto no plano de projecao              */
      o1 = 3, o2 = 1, o3 = 0.5,   /* Direcao do observador                   */
      O1 = 3, O2 = 1, O3 = 0.5;   /* Copias de o1, o2, o3                    */

/* ------------------------------------------------------------------------- */

#define SQR(x)  ((x)*(x))

#define DISTANCIA(a, b, c, x, y, z) sqrt(SQR(a-x) + SQR(b-y) + SQR(c-z))

#define PROD_VET(V1, V2, V3, a1, a2, a3, b1, b2, b3) V1 = a2*b3-a3*b2; \
				   V2 = a3*b1-a1*b3; V3 = a1*b2-a2*b1

#define NORMA(V1, V2, V3)  (sqrt((V1)*(V1) + (V2)*(V2) + (V3)*(V3)))

/* ------------------------------------------------------------------------- */

int IniciaModoGrafico(void) {

/* Inicializa o modo grafico. Retorna 1 no caso de sucesso ou 0 em caso
   contrario. */

  int gm, gd = DETECT;

  registerbgidriver(EGAVGA_driver);
  initgraph(&gd, &gm, "");
  return (graphresult() == grOk);
}

/* ------------------------------------------------------------------------ */

void DeterminaPlanoDeProjecao(void) {

/* Determina o  plano de  projecao ax + by + cz + d = 0, um ponto do plano
   (P1, P2, P3) e 2 vetores tangentes ao plano (u1, u2, u3) e (v1, v2, v3)
   perpendiculares entre si. */

  float N_o, prod_aux, t, N_u, N_v, compr_aux = 10;

  N_o = NORMA(o1, o2, o3);
  a = o1; b = o2; c = o3;
  o1 = o1/N_o; o2 = o2/N_o; o3 = o3/N_o;
  x0 = o1*dist_obs; y0 = o2*dist_obs; z0 = o3*dist_obs;
  P1 = o1*dist_plano; P2 = o2*dist_plano; P3 = o3*dist_plano;
  d = -a*P1 - b*P2 - c*P3;

  /* Determinacao de 2 vetores ortogonais no plano de projecao */

  if (o1 != 0) {
    t = 1/(1 - compr_aux*c/(a*x0 + b*y0 + c*z0 + d));
    u1 = x0 + t*(P1 - x0) - P1;
    u2 = y0 + t*(P2 - y0) - P2;
    u3 = z0 + t*(P3 - z0 + compr_aux) - P3;
    N_u = NORMA(u1, u2, u3);
    u1 = u1/N_u; u2 = u2/N_u; u3 = u3/N_u;

    t = 1/(1 - compr_aux*b/(a*x0 + b*y0 + c*z0 + d));
    v1 = x0 + t*(P1 - x0) - P1;
    v2 = y0 + t*(P2 - y0 + compr_aux) - P2;
    v3 = z0 + t*(P3 - z0) - P3;
    N_v = NORMA(v1, v2, v3);
    v1 = v1/N_v; v2 = v2/N_v; v3 = v3/N_v;
  }

  else if (o2 != 0) {
    t = 1/(1 - compr_aux*c/(a*x0 + b*y0 + c*z0 + d));
    u1 = x0 + t*(P1 - x0) - P1;
    u2 = y0 + t*(P2 - y0) - P2;
    u3 = z0 + t*(P3 - z0 + compr_aux) - P3;
    N_u = NORMA(u1, u2, u3);
    u1 = u1/N_u; u2 = u2/N_u; u3 = u3/N_u;

    t = 1/(1 - compr_aux*a/(a*x0 + b*y0 + c*z0 + d));
    v1 = x0 + t*(P1 - x0 + compr_aux) - P1;
    v2 = y0 + t*(P2 - y0) - P2;
    v3 = z0 + t*(P3 - z0) - P3;
    N_v = NORMA(v1, v2, v3);
    v1 = v1/N_v; v2 = v2/N_v; v3 = v3/N_v;
  }

  else if (o3 != 0) {
    t = 1/(1 - compr_aux*a/(a*x0 + b*y0 + c*z0 + d));
    u1 = x0 + t*(P1 - x0 + compr_aux) - P1;
    u2 = y0 + t*(P2 - y0) - P2;
    u3 = z0 + t*(P3 - z0) - P3;
    N_u = NORMA(u1, u2, u3);
    u1 = u1/N_u; u2 = u2/N_u; u3 = u3/N_u;

    t = 1/(1 - compr_aux*b/(a*x0 + b*y0 + c*z0 + d));
    v1 = x0 + t*(P1 - x0) - P1;
    v2 = y0 + t*(P2 - y0 + compr_aux) - P2;
    v3 = z0 + t*(P3 - z0) - P3;
    N_v = NORMA(v1, v2, v3);
    v1 = v1/N_v; v2 = v2/N_v; v3 = v3/N_v;
  }

  prod_aux = u1*v1 + u2*v2 + u3*v3;
  v1 = v1 - prod_aux*u1;
  v2 = v2 - prod_aux*u2;
  v3 = v3 - prod_aux*u3;
}

/* ------------------------------------------------------------------------- */

void ProjetaPonto(float x1, float y1, float z1, float *x, float *y) {

/* Determina as coordenadas do ponto (x1, y1, z1) no plano de projecao. Deve
   ser feita antes uma chamada a DeterminaPlanoDeProjecao(). */

  float q1, q2, q3, t;

  t = 1/(1 - (a*x1 + b*y1 + c*z1 + d)/(a*x0 + b*y0 + c*z0 + d));
  q1 = x0 + t*(x1 - x0);
  q2 = y0 + t*(y1 - y0);
  q3 = z0 + t*(z1 - z0);
  *x = (q1 - P1)*v1 + (q2 - P2)*v2 + (q3 - P3)*v3;
  *y = (q1 - P1)*u1 + (q2 - P2)*u2 + (q3 - P3)*u3;
}

/* ------------------------------------------------------------------------- */

void DesenhaEixos(int tam, int cor) {

/* Desenha os eixos x, y, z usando tamanho e cor especificados */

  float X0, Y0, X, Y;
  int cor_aux, x_x, x_y, y_x, y_y, z_x, z_y;

  cor_aux = getcolor();
  setcolor(cor);
  ProjetaPonto(0, 0, 0, &X0, &Y0);
  ProjetaPonto(tam, 0, 0, &X, &Y);
  line((int) (A + X0*TAM), (int) (B - Y0*TAM),
       (int) (A + X*TAM), (int) (B - Y*TAM));
  x_x = (int) (A + X*TAM); x_y = (int) (B - Y*TAM);
  ProjetaPonto(0, tam, 0, &X, &Y);
  line((int) (A + X0*TAM), (int) (B - Y0*TAM),
       (int) (A + X*TAM), (int) (B - Y*TAM));
  y_x = (int) (A + X*TAM); y_y = (int) (B - Y*TAM);
  ProjetaPonto(0, 0, tam, &X, &Y);
  line((int) (A + X0*TAM), (int) (B - Y0*TAM),
       (int) (A + X*TAM), (int) (B - Y*TAM));
  z_x = (int) (A + X*TAM); z_y = (int) (B - Y*TAM);

  setcolor(cor_aux);
  outtextxy(x_x, x_y, "x");
  outtextxy(y_x, y_y, "y");
  outtextxy(z_x, z_y, "z");
}

/* ------------------------------------------------------------------------- */

void DesenhaPonto(float x, float y, float z) {

/* Desenha a projecao do ponto (x, y, z) no plano de projecao. Deve
   ser feita antes uma chamada a DeterminaPlanoDeProjecao(). */

  float X, Y, q1, q2, q3, t;

  t = 1/(1 - (a*x + b*y + c*z + d)/(a*x0 + b*y0 + c*z0 + d));
  q1 = x0 + t*(x - x0);
  q2 = y0 + t*(y - y0);
  q3 = z0 + t*(z - z0);
  X = (q1 - P1)*v1 + (q2 - P2)*v2 + (q3 - P3)*v3;
  Y = (q1 - P1)*u1 + (q2 - P2)*u2 + (q3 - P3)*u3;
  putpixel((int) (A + X*TAM), (int) (B - Y*TAM), getcolor());
}

/* ------------------------------------------------------------------------- */

void DesenhaPontoCor(float x, float y, float z, int cor) {

/* Desenha a projecao do ponto (x, y, z) no plano de projecao. Deve
   ser feita antes uma chamada a DeterminaPlanoDeProjecao(). E' usa-
   da a cor fornecida. */

  float X, Y, q1, q2, q3, t;

  t = 1/(1 - (a*x + b*y + c*z + d)/(a*x0 + b*y0 + c*z0 + d));
  q1 = x0 + t*(x - x0);
  q2 = y0 + t*(y - y0);
  q3 = z0 + t*(z - z0);
  X = (q1 - P1)*v1 + (q2 - P2)*v2 + (q3 - P3)*v3;
  Y = (q1 - P1)*u1 + (q2 - P2)*u2 + (q3 - P3)*u3;
  putpixel((int) (A + X*TAM), (int) (B - Y*TAM), cor);
}

/* ------------------------------------------------------------------------- */

void DesenhaSuperficie5(float (*f) (float, float), float (*g) (float, float),
     float (*h) (float, float), float umin, float umax,
     float vmin, float vmax, int cor1, int cor2) {

/* Desenha o grafico da funcao (x, y, f(x, y)) usando o algoritmo conhecido
como "horizonte flutuante" */

  float X, Y, u, v;
  int horizonte_max[2000], horizonte_min[2000], maxx, i, x, y;

  maxx = getmaxx();
  for (i = 0; i <= maxx; i++) {
    horizonte_max[i] = 9999;
    horizonte_min[i] = -9999;
  }

  for (u = umax; u > umin; u -= (umax - umin)/nu) {
    for (v = vmin; v < vmax; v += inc) {
      ProjetaPonto(f(u, v), g(u, v), h(u, v), &X, &Y);
      x = (int) (A + X*TAM); y = (int) (B - Y*TAM);
      if (x >= 0 && x <= maxx) {
	if (y < horizonte_max[x]) {
	  horizonte_max[x] = y;
	  DesenhaPontoCor(f(u, v), g(u, v), h(u, v), cor1);
	}
	if (y > horizonte_min[x]) {
	  horizonte_min[x] = y;
	  DesenhaPontoCor(f(u, v), g(u, v), h(u, v), cor2);
	}
      }
    }

    v = vmax;
    ProjetaPonto(f(u, v), g(u, v), h(u, v), &X, &Y);
    x = (int) (A + X*TAM); y = (int) (B - Y*TAM);
    if (x >= 0 && x <= maxx) {
      if (y < horizonte_max[x]) {
	horizonte_max[x] = y;
	DesenhaPontoCor(f(u, v), g(u, v), h(u, v), cor1);
      }
      if (y > horizonte_min[x]) {
	horizonte_min[x] = y;
	DesenhaPontoCor(f(u, v), g(u, v), h(u, v), cor2);
      }
    }
  }

  u = umin;
  for (v = vmin; v < vmax; v += inc) {
    ProjetaPonto(f(u, v), g(u, v), h(u, v), &X, &Y);
    x = (int) (A + X*TAM); y = (int) (B - Y*TAM);
    if (x >= 0 && x <= maxx) {
      if (y < horizonte_max[x]) {
	horizonte_max[x] = y;
	DesenhaPontoCor(f(u, v), g(u, v), h(u, v), cor1);
      }
      if (y > horizonte_min[x]) {
	horizonte_min[x] = y;
	DesenhaPontoCor(f(u, v), g(u, v), h(u, v), cor2);
      }
    }
  }

  v = vmax;
  ProjetaPonto(f(u, v), g(u, v), h(u, v), &X, &Y);
  x = (int) (A + X*TAM); y = (int) (B - Y*TAM);
  if (x >= 0 && x <= maxx) {
    if (y < horizonte_max[x]) {
      horizonte_max[x] = y;
      DesenhaPontoCor(f(u, v), g(u, v), h(u, v), cor1);
    }
    if (y > horizonte_min[x]) {
      horizonte_min[x] = y;
      DesenhaPontoCor(f(u, v), g(u, v), h(u, v), cor2);
    }
  }
}

/* ------------------------------------------------------------------------- */

/*** FIM DE GRAFICOS.C ***/
